from functools import wraps
from .utils import load_from_str_or_buf
import json as json_
from os import makedirs
from os.path import dirname

@wraps(json_.load)
def load(path_or_buf, **kwargs):
    return json_.load(fp=load_from_str_or_buf(path_or_buf), **kwargs)

@wraps(json_.dump)
def dump(obj, path_or_buf, **kwargs):
    if isinstance(path_or_buf, str):
        makedirs(dirname(path_or_buf), exist_ok=True)
        with open(path_or_buf, 'w', encoding='utf-8') as fp:
            return dump(obj, path_or_buf=fp, **kwargs)
    
    return json_.dump(obj=obj, fp=path_or_buf, **kwargs)

def __getattr__(x):
    return getattr(json_, x)

import pandas as pd

def generate_json_schema(data):
    def infer_schema(value):
        if isinstance(value, dict):
            properties = {}
            required = []
            for k, v in value.items():
                properties[k] = infer_schema(v)
                required.append(k)
            return {
                "type": "object",
                "properties": properties,
                "required": required
            }
        elif isinstance(value, list):
            if value:
                item_schemas = [infer_schema(item) for item in value]
                # Simplify if all item schemas are the same
                if all(schema == item_schemas[0] for schema in item_schemas):
                    return {
                        "type": "array",
                        "items": item_schemas[0]
                    }
                else:
                    return {
                        "type": "array",
                        "items": {
                            "anyOf": item_schemas
                        }
                    }
            else:
                # Empty array; cannot determine item schema
                return {
                    "type": "array",
                    "items": {}
                }
        else:
            # Primitive types
            if isinstance(value, str):
                return {"type": "string"}
            elif isinstance(value, bool):
                return {"type": "boolean"}
            elif isinstance(value, int):
                return {"type": "integer"}
            elif isinstance(value, float):
                return {"type": "number"}
            elif value is None:
                return {"type": "null"}
            else:
                return {}

    return infer_schema(data)

def json_to_table(json_schema, json_data):
    def flatten_json(data, schema, parent_key='', parent_rows=None):
        if parent_rows is None:
            parent_rows = [{}]

        data_type = schema.get('type')

        if data_type == 'object':
            properties = schema.get('properties', {})
            for key, subschema in properties.items():
                value = data.get(key)
                new_key = f"{parent_key}.{key}" if parent_key else key

                new_parent_rows = []
                for parent_row in parent_rows:
                    if value is not None:
                        sub_rows = flatten_json(value, subschema, new_key, [parent_row.copy()])
                        new_parent_rows.extend(sub_rows)
                    else:
                        parent_row[new_key] = None
                        new_parent_rows.append(parent_row)
                parent_rows = new_parent_rows

        elif data_type == 'array':
            items_schema = schema.get('items')
            new_parent_rows = []
            if isinstance(data, list):
                for item in data:
                    for parent_row in parent_rows:
                        sub_rows = flatten_json(item, items_schema, parent_key, [parent_row.copy()])
                        new_parent_rows.extend(sub_rows)
                parent_rows = new_parent_rows
            else:
                # Data is not a list but schema says it should be
                pass  # Handle error or treat as empty list
        else:
            # Primitive type
            for parent_row in parent_rows:
                parent_row[parent_key] = data

        return parent_rows

    rows = flatten_json(json_data, json_schema)
    df = pd.DataFrame(rows)
    return df


import re
from time import sleep
import numpy.random as random

import html
import urllib.request
import urllib.parse

agent = {
    'User-Agent':
    "Mozilla/4.0 (\
    compatible;\
    MSIE 6.0;\
    Windows NT 5.1;\
    SV1;\
    .NET CLR 1.1.4322;\
    .NET CLR 2.0.50727;\
    .NET CLR 3.0.04506.30\
    )"}

CLASS_SELECTOR = re.compile(r'(?s)class="(?:t0|result-container)">(.*?)<')
BASE_URL = "http://translate.google.com/m?tl={target}&sl={source}&q={text}"

def __translate__(to_translate, to_language='auto', from_language='auto'):
    sleep(random.uniform(.5,.7))
    to_translate = urllib.parse.quote(to_translate)
    link = BASE_URL.format(source=from_language, 
                            target=to_language, 
                            text=to_translate)
    # print(link)
    request = urllib.request.Request(link, headers=agent)
    raw_data = urllib.request.urlopen(request).read()
    data = raw_data.decode("utf-8")
    re_result = CLASS_SELECTOR.findall(data)
    result = '' if not re_result else html.unescape(re_result[0])
    return result

def __call__(to_translate, to_language='auto', from_language='auto'):
    return __translate__(to_translate, to_language, from_language)

def bulk(*strs, 
        input_lang='auto', 
        output_lang='auto', 
        delimiter=";\n") -> list[str]:
    
    s = delimiter.join(strs)
    o = __translate__(s, to_language=input_lang, from_language=output_lang)
    return o.split(delimiter)

import sys

class Module(sys.modules[__name__].__class__):
    def __call__(self, to_translate, to_language='auto', from_language='auto'):
        return __call__(to_translate, to_language, from_language)

sys.modules[__name__].__class__ = Module


from tqdm.contrib.concurrent import ensure_lock, thread_map as _thread_map, process_map as _process_map
from tqdm.auto import tqdm as tqdm_auto
from functools import wraps
from os import cpu_count
from concurrent.futures.process import BrokenProcessPool
import sys

def _executor(PoolExecutor, fs, **tqdm_kwargs):
    """
    Implementation of `thread_run_all` and `process_run_all`.

    Parameters
    ----------
    tqdm_class  : [default: tqdm.auto.tqdm].
    max_workers  : [default: min(32, cpu_count() + 4)].
    chunksize  : [default: 1].
    lock_name  : [default: "":str].
    """
    kwargs = tqdm_kwargs.copy()

    if "total" not in kwargs:
        kwargs["total"] = len(fs)

    tqdm_class = kwargs.pop("tqdm_class", tqdm_auto)
    max_workers = kwargs.pop("max_workers", min(32, cpu_count() + 4))
    chunksize = kwargs.pop("chunksize", 1)
    lock_name = kwargs.pop("lock_name", "")
    with ensure_lock(tqdm_class, lock_name=lock_name) as lk:
        pool_kwargs = {'max_workers': max_workers}
        sys_version = sys.version_info[:2]
        if sys_version >= (3, 7):
            # share lock in case workers are already using `tqdm`
            pool_kwargs.update(initializer=tqdm_class.set_lock, initargs=(lk,))

        with PoolExecutor(**pool_kwargs) as ex:
            futures = [ex.submit(func) for func in fs]
            return [f.result() for f in tqdm_class(futures, **kwargs)]

def _process_run_all(fs, **tqdm_kwargs):
    from concurrent.futures import ProcessPoolExecutor
    if "lock_name" not in tqdm_kwargs:
        tqdm_kwargs = tqdm_kwargs.copy()
        tqdm_kwargs["lock_name"] = "mp_lock"
    return _executor(ProcessPoolExecutor, fs, **tqdm_kwargs)


BROKEN_POOL_WIN_EXCEPTION = \
    BrokenProcessPool("If you're on Windows, the process pool has to be instantiated inside 'if __name__ == '__main__': ...'")

@wraps(_process_run_all)
def process_run_all(*args, **kwargs):
    try:
        return _process_run_all(*args, **kwargs)
    except BrokenProcessPool:
        raise BROKEN_POOL_WIN_EXCEPTION

def thread_run_all(fs, **tqdm_kwargs):
    from concurrent.futures import ThreadPoolExecutor
    if "lock_name" not in tqdm_kwargs:
        tqdm_kwargs = tqdm_kwargs.copy()
        tqdm_kwargs["lock_name"] = "thread_lock"
    return _executor(ThreadPoolExecutor, fs, **tqdm_kwargs)

thread_map = _thread_map

@wraps(_process_map)
def process_map(*args, **kwargs):
    try:
        return _process_map(*args, **kwargs)
    except BrokenProcessPool:
        raise BROKEN_POOL_WIN_EXCEPTION

def seq_run_all(fs, **tqdm_kwargs):
    kwargs = tqdm_kwargs.copy()

    if "total" not in kwargs:
        kwargs["total"] = len(fs)
        
    tqdm_class = kwargs.pop("tqdm_class", tqdm_auto)

    return list(tqdm_class([f() for f in fs], **kwargs))

def seq_map(fs, *iterables, **tqdm_kwargs):
    kwargs = tqdm_kwargs.copy()

    if "total" not in kwargs:
        kwargs["total"] = len(iterables)

    tqdm_class = kwargs.pop("tqdm_class", tqdm_auto)


    return list(tqdm_class(map(fs, *iterables), **kwargs))


from lingua import Language

language_codes = {
    "AFRIKAANS": 'af',
    "ALBANIAN": 'sq',
    "ARABIC": 'ar',
    "ARMENIAN": 'hy',
    "AZERBAIJANI": 'az',
    "BASQUE": 'eu',
    "BELARUSIAN": 'be',
    "BENGALI": 'bn',
    "BOKMAL": 'nb',
    "BOSNIAN": 'bs',
    "BULGARIAN": 'bg',
    "CATALAN": 'ca',
    "CHINESE": 'zh',
    "CROATIAN": 'hr',
    "CZECH": 'cs',
    "DANISH": 'da',
    "DUTCH": 'nl',
    "ENGLISH": 'en',
    "ESPERANTO": 'eo',
    "ESTONIAN": 'et',
    "FINNISH": 'fi',
    "FRENCH": 'fr',
    "GANDA": 'lg',
    "GEORGIAN": 'ka',
    "GERMAN": 'de',
    "GREEK": 'el',
    "GUJARATI": 'gu',
    "HEBREW": 'he',
    "HINDI": 'hi',
    "HUNGARIAN": 'hu',
    "ICELANDIC": 'is',
    "INDONESIAN": 'id',
    "IRISH": 'ga',
    "ITALIAN": 'it',
    "JAPANESE": 'ja',
    "KAZAKH": 'kk',
    "KOREAN": 'ko',
    "LATIN": 'la',
    "LATVIAN": 'lv',
    "LITHUANIAN": 'lt',
    "MACEDONIAN": 'mk',
    "MALAY": 'ms',
    "MAORI": 'mi',
    "MARATHI": 'mr',
    "MONGOLIAN": 'mn',
    "NYNORSK": 'nn',
    "PERSIAN": 'fa',
    "POLISH": 'pl',
    "PORTUGUESE": 'pt',
    "PUNJABI": 'pa',
    "ROMANIAN": 'ro',
    "RUSSIAN": 'ru',
    "SERBIAN": 'sr',
    "SHONA": 'sn',
    "SLOVAK": 'sk',
    "SLOVENE": 'sl',
    "SOMALI": 'so',
    "SOTHO": 'st',
    "SPANISH": 'es',
    "SWAHILI": 'sw',
    "SWEDISH": 'sv',
    "TAGALOG": 'tl',
    "TAMIL": 'ta',
    "TELUGU": 'te',
    "THAI": 'th',
    "TSONGA": 'ts',
    "TSWANA": 'tn',
    "TURKISH": 'tr',
    "UKRAINIAN": 'uk',
    "URDU": 'ur',
    "VIETNAMESE": 'vi',
    "WELSH": 'cy',
    "XHOSA": 'xh',
    "YORUBA": 'yo',
    "ZULU": 'zu'
}


import pandas as pandas_
from pandas.core.series import Series as Series_
from copy import copy
from functools import partial, wraps
from typing import TypeAlias, NewType, Type
import csv as csv_

original_to_csv = copy(pandas_.DataFrame.to_csv)

class DataFrame(pandas_.DataFrame): pass
del DataFrame

class Series(Series_): pass
del Series

formato_fundar = {
    'encoding': 'utf-8',
    'sep': ',',
    'quoting': csv_.QUOTE_ALL,
    'quotechar': '"',
    'lineterminator': '\n',
    'decimal': '.',
    'index': False,
    'float_format': '%.5f'
}

@wraps(original_to_csv)
def to_csv_patch(self, path_or_buf, **kwargs):
    return original_to_csv(self, path_or_buf, **(formato_fundar|kwargs))

pandas_.DataFrame.to_csv = to_csv_patch
setattr(Series_, 'has', lambda self, x: self.map(lambda y: x in y))
setattr(Series_, 'has_safe', lambda self, x: self.map(lambda y: False if '__contains__' not in dir(y) else x in y))

def __getattr__(name):
    return getattr(pandas_, name)


import io
import os
from operator import contains
from functools import reduce, partial
from typing import NewType, Callable, TypeVar, Protocol, Generic, Optional
from datetime import datetime

class classproperty(property):
    """Utilidad para crear propiedades de clase"""

    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()

# =============================================================================================

T = TypeVar('T')

class Indexable(Protocol[T]):
    def __getitem__(self, index: int) -> T: ...

class number(object): ...
number = int|float

# =============================================================================================

def compose2(f, g):
    return lambda *args, **kwargs: f(g(*args, **kwargs))

def compose(*fs):
    return reduce(compose2, fs)

def has(x):
    """Curried version of operator.contains"""
    return lambda y: contains(y, x)

def xmax(**kwargs):
    """
    Curried version of max.
    """
    return lambda x: max(x, **kwargs)

def call(f, *args, **kwargs):
    return f(*args, **kwargs)

def callx(*args, **kwargs):
    return lambda f: call(f, *args, **kwargs)

def access(i):
    return lambda x: x[i]

# https://hackage.haskell.org/package/base-4.19.1.0/docs/Prelude.html#v:fst
fst = access(0)

# https://hackage.haskell.org/package/base-4.19.1.0/docs/Prelude.html#v:snd
snd = access(1)

def identity_single_argument(x):
    return x

def identity(*args):
    return args

def is_empty(x):
    return False if x else True

def _apply_to(f, i, x):
    return type(x)(f(y) if (j == i) else y for j,y in enumerate(x))

def apply_to(f, i):
    """
    Applies a function to the ith element of an indexable object.
    Leaves the rest just as is.
    """
    return lambda x: _apply_to(f, i, x)

# https://hackage.haskell.org/package/base-4.19.1.0/docs/Control-Arrow.html#v:first
first = partial(apply_to, i=0)

# https://hackage.haskell.org/package/base-4.19.1.0/docs/Control-Arrow.html#v:second
second = partial(apply_to, i=1)

A = TypeVar('A')
def group_by(predicate: Callable[[A, A], bool], elements: list[A]) -> list[list[A]]:
    if not elements:
        return []
    
    def go(groups, x):
        if not groups:
            return [[x]]
        
        last_group = groups[-1]
        if predicate(x, last_group[0]):
            groups[-1] = last_group + [x]
        else:
            groups.append([x])
        return groups

    groups = []
    for element in elements:
        groups = go(groups, element)
    return groups

def _flatten(lst, flat_list):
    for item in lst:
        if isinstance(item, list):
            _flatten(item, flat_list)
        else:
            flat_list.append(item)

    return flat_list

def flatten(nested_list):
    return _flatten(nested_list, [])


def groupby(predicate, elements):
    match_count = 0
    
    for x in elements:
        if predicate(x):
            match_count += 1
    
    left = [None for _ in range(match_count)]
    right = [None for _ in range(len(elements) - match_count)]

    i, j = 0, 0
    for x in elements:
        if predicate(x):
            left[i] = x
            i += 1
        else:
            right[j] = x
            j += 1
    
    return left, right
    
# =============================================================================================

class staticproperty(property):
    """Utilidad para crear propiedades estáticas (que no tomen el puntero a self)"""

    def __get__(self, cls, owner):
        return staticmethod(self.fget).__get__(None, owner)()

class SingletonMeta(type):
    """
    Fuente: github.com/datos-Fundar/fundartools
    Metaclase que implementa el patrón de singleton.
    Las clases que la heredan no pueden ser instanciadas más de una vez.
    Provee un método 'get_instance' que es heredado, el cual:
    - Si la clase no está instanciada, la crea.
    - Si está instanciada, la devuelve.
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
            return instance

        raise RuntimeError("Class already instantiated. Use get_instance()")

    def get_instance(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]
    
class Singleton(metaclass=SingletonMeta):
    def __init_subclass__(cls) -> None:
        cls.instance = classproperty(lambda _: cls.get_instance())
    
# =============================================================================================

_T_co = TypeVar('_T_co')
_T = TypeVar('_T')

class SupportsNext(Protocol[_T_co]):
    def __next__(self) -> _T_co: ...


class SupportsIter(Protocol[_T_co]):
    def __iter__(self) -> _T_co: ...

class Attribute(Generic[_T]): ...

class Final(Generic[_T]):
    """Indica que la variable no puede ser reasignada, pero si puede mutar."""


class Mutable(Generic[_T]):
    """Indica que la variable puede ser reasignada y mutar"""


class Inmutable(Generic[_T]):
    """Indica que la variable no puede ser reasignada ni mutar"""

def getattrc(attr: str):
    """Versión currificada de 'getattr'"""
    return lambda obj: getattr(obj, attr)

# =============================================================================================

Integer = TypeVar('Integer', int, int)
Float = TypeVar('Float', float, float)
Number = TypeVar('Number', int, float)

class Range(tuple[Number, Number], Generic[Number]): ...

class Max(tuple[T, ...]): ...
class Min(tuple[T, ...]): ...

class FloatRange(Range[float]): ...
class IntRange(Range[int]): ...

class UnboundedFloat(float): ...

class BoundedFloat(tuple[Float, Float]): ...
class BoundedInt(tuple[Integer, Integer]): ...

class Mapper(Callable[[Range], Callable[[Range, UnboundedFloat], BoundedFloat]]): ...

def mapper(ostart, ostop) -> Mapper:
    return lambda value, istart, istop: map_value(value, istart, istop, ostart, ostop)

def map_value(value, istart, istop, ostart, ostop):
    return ostart + (ostop - ostart) * ((value - istart) / (istop - istart))

# =============================================================================================

def load_from_str_or_buf(input_data) -> io.BytesIO:
    match input_data:
        case string if isinstance(input_data, str):
            if not os.path.exists(input_data):
                raise FileNotFoundError(f'File {string} not found.')
            if not os.path.isfile(string):
                raise ValueError("Input is a folder, not a file.")
            
            with open(string, 'rb') as file:
                return io.BytesIO(file.read())
            
        case buffer if isinstance(input_data, (io.BytesIO, io.StringIO)):
            return buffer
        case _:
            raise TypeError("Unsupported input type. Please provide a valid file path or buffer.")
        
# =============================================================================================

def apply(f):
    """
    ...args :: tuple[varargs, varkwargs]
    y :: f(..args): Any
    f -> ...args -> y | tuple[...args, y]
    """
    def apply_to(*args, **kwargs):
        n = len(args)
        m = len(kwargs)
        if n+m == 0:
            return f()
        
        if n+m == n:
            if n == 1:
                x = args[0]
                return (x, f(x))
            else:
                return (args, f(*args))
        
        if n+m == m:
            return (kwargs, f(**kwargs))
        
        return (args, kwargs, f(*args, **kwargs))
    return apply_to

# =============================================================================================

from typing import Iterable, Union, Type, Dict

def text_wrap(s: str, max_length=100) -> str:
    if len(s) >= max_length:
        return s[:max_length]+'\n'+text_wrap(s[max_length:], max_length)
    return s

def is_subscrevable(x: object|type) -> bool: # Raises Exception^AttributeError
    if not isinstance(x, type):
        return is_subscrevable(type(x))
    try:
        x.__getitem__
        return True
    except Exception as e:
        match e:
            case AttributeError():
                return False
            case _:
                raise e

def internal_type_of_iterable(s: Iterable) -> type|Union[type]:
    missing_first = object()
    iterator = iter(s)
    first_element = next(iterator, missing_first)
    
    if first_element is missing_first:
        raise ValueError('Empty iterable')

    internal_type = type(first_element)
    for t in map(type, iterator):
        internal_type |= t

    return internal_type

def _typeof_iterable(s: Iterable) -> tuple[Type[Iterable], type|Union[type]]:
    base_type = type(s)
    internal_type = internal_type_of_iterable(s)
    
    if not is_subscrevable(base_type):
        print(base_type)
        return None
    return base_type, internal_type

def typeof_iterable(s: Iterable) -> Type[Type]:
    base_type, internal_type = _typeof_iterable(s)
    return base_type[internal_type]

def _typeof_dict(d: dict) -> tuple[Type[dict], type|Union[type]]:
    keys = internal_type_of_iterable(d.keys())
    values = internal_type_of_iterable(d.values())
    return keys, values

def typeof_dict(d: dict) -> Type[Dict[type|Union[type], type|Union[type]]]:
    keys, values = _typeof_dict(d)
    return dict[keys, values]

def isiterable(s: object) -> bool:
    try:
        iter(s)
        return True
    except TypeError:
        return False

def split(condition):
    def _split(xs):
        successful, failed = [], []
        for x in xs:
            (successful if condition(x) else failed).append(x)
        return successful, failed
    return _split

from inspect import signature

def argumentsOf(f): return len(signature(f).parameters)

def curry(f: callable):
    """
    Currifica una funcion de der. a izq.
    Por ejemplo:

    curry(lambda a,b: a/b) === lambda b: lambda a: a/be
    """
    def curried(*args):
        if len(args) == argumentsOf(f): return f(*args)
        else: return lambda x: curried(x, *args)
    return curried

# =============================================================================================

class Negate(object):
    def __matmul__(self, other):
        return lambda *args, **kwargs: not other(*args, **kwargs)

class negate(): ...
del negate

negate = Negate()

# =============================================================================================

def print_id(x):
    print(x)
    return x

# =============================================================================================

class now:
    # noinspection PyMethodParameters
    @staticproperty
    def string():
        return datetime.now().strftime('%d-%m-%y_%H%M%S')
    
    def format(fmt):
        return datetime.now().strftime(fmt)

# ============================================================================================

class MethodMapping(dict):
    """Diccionario que asocia claves a funciones. Sirve para crear selectores de estrategia."""

    def __register__(self, key, f):
        self[key] = f
        return f

    def register(self, key_or_alias):
        match key_or_alias:
            case alias if isinstance(key_or_alias, str):
                return lambda f: self.__register__(alias, f)
            case function if callable(key_or_alias):
                return self.__register__(function.__name__, function)
            
    def __getattr__(self, k):
        return self[k]

# ============================================================================================   

def throw(ex):
    raise ex

class Placeholder:    
    def __init_subclass__(cls) -> None:
        cls.__init__ = lambda *_, **__: throw(TypeError(f"'{cls.__name__}' is a placeholder and cannot be instantiated."))
        cls.__init_subclass__ = lambda *_, **__: throw(TypeError(f"'{cls.__name__}' is a placeholder and cannot be subtyped."))

# ============================================================================================

def search_downwards(name: str, start_path: str, max_depth: int, current_depth: int=0) -> Optional[str]:
    if current_depth > max_depth:
        return None
    
    for root, dirs, files in os.walk(start_path):
        if name in files or name in dirs:
            return os.path.join(root, name)
        
        if current_depth + 1 > max_depth:
            break
        
        for d in dirs:
            result = search_downwards(name, os.path.join(root, d), max_depth, current_depth + 1)
            if result:
                return result

    return None

def search_upwards(name: str, start_path: str, max_up_depth: int, max_down_depth: int, current_up_depth: int=0) -> Optional[str]:
    if current_up_depth > max_up_depth:
        return None
    
    result = search_downwards(name, start_path, max_down_depth)
    if result:
        return result
    
    parent_dir = os.path.abspath(os.path.join(start_path, '..'))
    
    if parent_dir == start_path:
        return None
    
    return search_upwards(name, parent_dir, max_up_depth, max_down_depth, current_up_depth + 1)

def find_file(name: str, path: str, max_up_depth: int=3, max_down_depth: int=3, throw_error: bool=False) -> Optional[str]:
    result = search_downwards(name, path, max_down_depth)

    if result:
        # get absolute path
        result = os.path.abspath(result)
        return result
    

    result = search_upwards(name, path, max_up_depth, max_down_depth)
    if result:
        # get absolute path
        result = os.path.abspath(result)
        return result
    
    if throw_error:
        raise FileNotFoundError(f"File '{name}' not found.")
    
def split_path_recursively(path: str) -> list:
    if path == os.path.sep:
        return []
    
    head, tail = os.path.split(path)
    
    if head == '' or head == path:
        return [tail]
    
    return split_path_recursively(head) + [tail]



from pandas import option_context
from IPython.display import display

show_full_dataframe_context = option_context('display.max_rows', None, 'display.max_columns', None)

def show_full_dataframe(df):
    with show_full_dataframe_context:
        display(df)




from typing import Callable, Iterable, TypeVar, Any
from .pools import *

A = TypeVar('A')
B = TypeVar('B')
class Executor:
    run_all: Callable
    run_map: Callable

    @classmethod
    def execute(cls, fs: Iterable[Callable[[], A]], **tqdm_kwargs) -> list[A]:
        """
        Executes a list of functions with void arguments and returns their results in order.
        """
        return cls.run_all(fs, **tqdm_kwargs)

    @classmethod
    def map(cls, fn: Callable[[A], B], *iterables: A, **tqdm_kwargs) -> list[B]:
        """
        Maps and executes a function over variadic arguments and returns the results in order.
        """
        return cls.run_map(fn, *iterables, **tqdm_kwargs)
    
class Multithread(Executor):
    """
    Executes actions using multiple threads.
    """

    run_all = thread_run_all
    run_map = thread_map

class Multiprocess(Executor):
    """
    Executes actions using multiple processes.
    """
    run_all = process_run_all
    run_map = process_map

class Sequentially(Executor):
    """
    Executes actions sequentially (single-thread, single-process).
    """
    run_all = seq_run_all
    run_map = seq_map


from .utils import mapper, Mapper, Range, Indexable, BoundedInt, BoundedFloat, Max
from typing import TypeVar

A = TypeVar('A')
B = TypeVar('B')

N = TypeVar('N', float, float)
M = TypeVar('M', float, float)

normaliser: Mapper[[Range[0, 1]], ...] = mapper(0, 1)
"""
Normaliza un valor en el rango [0, 1].

Uso:
    normaliser(x, n, m) -> y
    Donde:
        x es es un numero que está entre [n, m]
        y es un número que está entre [0, 1]
"""

inverse_normaliser: Mapper[[Range[1, 0]], ...] = mapper(1, 0)
"""
Normaliza un valor en el rango [1, 0].

Uso:
    inverse_normaliser(x, n, m) -> y
    Donde:
        x es es un numero que está entre [n, m]
        y es un número que está entre [1, 0]
"""

def levenshtein_distance(a: Indexable[A], b: Indexable[A]) -> BoundedInt[0, Max[N, M]]:
    """
    Calcula la distancia de Levenshtein entre dos elementos. La distancia de Levenshtein es el número mínimo de
    operaciones requeridas para transformar una cadena en otra. 
    Las operaciones permitidas son inserción, eliminación y sustitución.

    Args:
        a: Indexable[A], donde len(a) = N
        b: Indexable[A], donde len(b) = M
        resultado: BoundedInt[0, max(N, M)]
    """
    n = len(a)
    m = len(b)

    matrix = [[0] * (m + 1) for _ in range(n + 1)]

    for i in range(n + 1):
        matrix[i][0] = i

    for j in range(m + 1):
        matrix[0][j] = j

    for i in range(1, n + 1):
        for j in range(1, m + 1):
            cost = 0 if a[i - 1] == b[j - 1] else 1
            matrix[i][j] = min(
                matrix[i - 1][j] + 1,  # deletion
                matrix[i][j - 1] + 1,  # insertion
                matrix[i - 1][j - 1] + cost  # substitution
            )

    # The bottom-right cell contains the minimum edit distance
    return matrix[n][m]

def normalized_levenshtein_similarity(a: Indexable[A], b: Indexable[A], mapper: Mapper[[Range[N, M]], ...] = None) -> BoundedFloat[N, M]:
    """
    Calcula la distancia de Levenshtein entre dos elementos y la normaliza en un rango.
    Si se usa el rango predefinido, el resultado estará en el rango [0, 1].
    Donde,
        0 significa que las cadenas son completamente distintas.
        1 significa que las cadenas son exactamente iguales.

    Args:
        a: Indexable[A]
        b: Indexable[B]
        mapper: Mapper[[N, M], ...], default = None
        resultado: BoundedFloat[N, M]
    """
    mapper = mapper or inverse_normaliser
    
    max_len = max(len(a), len(b))
    distance = levenshtein_distance(a,b)
    normalised_distance = mapper(distance, 0, max_len)
    
    return normalised_distance

def jaccard_similarity(a: set[A], b: set[A]) -> BoundedFloat[0, 1]:
    """
    Calcula la similitud de Jaccard entre dos conjuntos.
    La similitud de Jaccard es, intuitivamente, la proporción de elementos en común entre dos conjuntos.
    En particular, es el cociente entre los tamaños de la intersección y la unión de los conjuntos.

    El resultado es un número entre 0 y 1, donde:
        0 significa que los conjuntos son completamente disjuntos.
        1 significa que los conjuntos son exactamente iguales.

    Args:
        a: set[A]
        b: set[A]
        resultado: BoundedFloat[0, 1]
    """
    intersect_n = len(a & b)
    union_n = len(a | b)
    return intersect_n / union_n


from fundar.utils import staticproperty
from datetime import datetime, timedelta
from timeit import default_timer

class now:
    # noinspection PyMethodParameters
    @staticproperty
    def string():
        return datetime.now().strftime('%d-%m-%y_%H%M%S')

def stopwatch(f: callable, *args, **kwargs):
    start = default_timer()
    result = f(*args, **kwargs)
    elapsed_time = default_timer() - start
    return result, timedelta(seconds=elapsed_time)

def parse_time(time_string, timezone=None):
    original_datetime = datetime.fromisoformat(time_string.replace("Z", "+00:00"))
    if not timezone:
        return original_datetime
    
    timezone = pytz.timezone(timezone)
    return original_datetime.astimezone(timezone)

def parse_timex(timezone):
    return lambda time_string: parse_time(time_string, timezone=timezone)

parse_time_arg = parse_timex('America/Argentina/Buenos_Aires')

def format_time(time_object, format):
    return time_object.strftime(format)

def format_timex(format):
    return lambda time_object: format_time(time_object, format)

timeformat = format_timex("%d-%m-%y_%H%M%S")


try:
    import gwrappers
    from gwrappers.googledrive import *

    files_api = googledrive.instance.service.files()
except ImportError:
    raise ImportError("'gwrappers' isn't installed. Get it from github.com/datos-Fundar/gwrappers")


from setuptools import setup, find_packages

setup(
   version="1.0.1",
   name="fundar",
   author="Fundar",
   description="Private Python library.",
   packages=find_packages(),
   include_package_data=True,
   python_requires='>=3.10',
   setup_requires=['setuptools-git-versioning'],
   version_config={
       "dirty_template": "{tag}",
   }
)



from typing import Callable, Any, ClassVar
from collections.abc import Iterator, Iterable
from .utils import callx, flatten, access
from functools import reduce, wraps
from operator import eq as equals, add

class slicer:
    def __init__(self, stop, start=0, step=1):
        self._slice = slice(start, stop, step)

    @property
    def start(self) -> Any:
        return self._slice.start
    
    @property
    def step(self) -> Any:
        return self._slice.step
    
    @property
    def stop(self) -> Any:
        return self._slice.stop
    
    def __eq__(self, value: object, /) -> bool:
        return self._slice.__eq__(value)
    
    __hash__: ClassVar[None]  # type: ignore[assignment]

    def indices(self, *args) -> tuple[int, int, int]:
        return self._slice.indices(*args)
    
    def __call__(self, sliceable):
        return sliceable.__getitem__(self._slice)
    
    def __repr__(self):
        return repr(self._slice)
    
_missing = object()

class lista(list):
    """
    Subclase del primitivo 'list', con métodos adicionales para una manipulación de listas más funcional.
    En particular, convierte algunas operaciones lazy (map, filter, reduce) y las hace eager.
    """

    class _size(int):
        def __call__(self):
            return int(self)
        
    @classmethod
    def of(cls, *args):
        return lista(args)

    def map(*args):
        """
        Versión eager de map.
        """
        n = len(args)

        if n not in (1,2):
            raise ValueError(f'Expected 1 or 2 arguments, got {n}.')

        if n == 2:
            self, f = args
            return lista(map(f, self))
        
        if n == 1:
            f = args[0]
            return lambda xs: lista.map(xs, f)
        
    def reverse(self):
        return list(reversed(self))
    
    def unzip(self, *args):
        return list(zip(*self, *args))
    
    def flatmap(self, f):
        return list([item for sublist in self.map(f) for item in sublist])

    def filter(self, f):
        """
        Versión eager de filter.
        """
        return lista(filter(f, self))
    
    def reduce(self, f, initial=_missing):
        """
        Versión eager de reduce.
        """
        if initial is _missing:
            result = reduce(f, self)
        else:
            result = reduce(f, self, initial)

        if type(result) == list:
            return lista(result)
        return result
    
    foldl = reduce
    
    def foldr(self, f, initial=_missing):
        it = iter(self)
        if initial is _missing:
            try:
                value = next(it)
            except StopIteration:
                raise TypeError(
                    "foldr() of empty iterable with no initial value") from None
        else:
            value = initial

        for element in it:
            value = f(element, value)

        return value
    
    def safe_index(self, x, default=-1):
        """
        Devuelve el índice de x si está en la lista, o -1 en caso contrario.
        """
        try:
            return self.index(x)
        except ValueError:
            return default
        
    def enumerate(self):
        return lista(enumerate(self))
    
    def which(self, condition: Callable[..., bool]):
        return self.map(condition).enumerate().filter(access(1)).map(access(0))
    
    def zip(self, *args):
        """
        Versión eager de zip.
        """
        return lista(zip(self, *args))
    
    def zipmap(self, *fs, keep=True):
        if keep:
            return lista( self.zip(*(self.map(f) for f in fs)) )
        else:
            return lista( zip(*(self.map(f) for f in fs)) )
    
    def find(self, f, not_found=None):
        """
        Devuelve x tal que f(x) == True, si existe, o una lista vacía en caso contrario.
        """
        matches = self.filter(f)
        return matches[0] if matches else not_found
    
    def indexof(self, f):
        """
        Devuelve el índice de la primera ocurrencia de x tal que f(x) == True, si existe, o -1 en caso contrario.
        Para obtener el índice de un objeto particular, usar list.index.
        """
        result = self.find(f, not_found=None)
        return -1 if result is None else self.index(result) 
    
    def apply(*args, **kwargs):
        """
        Lista como functor aplicativo. (<*>)
        Sea L una lista de funciones [f1, f2, ..., fn] y x un valor, luego L.apply(x) es equivalente a [f1(x), f2(x), ..., fn(x)].
        """
        n = len(args)

        if n < 1:
            raise ValueError('lista.apply expected at least one argument.')
        
        match args:
            case [f] if callable(f):
                return lambda xs: lista(f(x, **kwargs) for x in xs)
            case [fs] if isinstance(fs, Iterable):
                return lambda xs: lista(f(x, **kwargs) for f in fs for x in xs)
            case [fs, ags] if isinstance(ags, Iterable):
                return lista(f(x) for f in fs for x in ags)
            case [fs, *ags]:
                return lista(f(x) for f in fs for x in ags)
    
    
    def _chain_together(self) -> callable:
        return lista.foldl(self, lambda f, g: (lambda *_args, **_kwargs: g(f(*_args, **_kwargs))))

    def chain(*args, **kwargs):
        n = len(args)

        if n < 1:
            raise ValueError('lista.chain expected at least one argument.')

        if n == 1:
            [x] = args
            if isinstance(x, Iterable):
                return lambda xs: lista._chain_together(x)(xs, **kwargs)
            elif callable(x):
                return lambda y: x(y)
        
        if isinstance(args[0], Iterable):
            [fs, *args] = args
            return lista._chain_together(fs)(*args, **kwargs)
        else:
            return lambda ys: lista._chain_together(args)(ys, **kwargs)
        
    def unique(self, comparator=equals):
        """
        Devuelve una lista con los elementos únicos de la lista original.
        """
        return lista(set(self))
    
    def nunique(self):
        """
        Devuelve el número de elementos únicos en la lista.
        """
        return len(self.unique())
    
    def flatten(self):
        return lista(flatten(self))
    
    def at(self, i):
        return self[i]
    
    def concat(self, *args):
        n = len(args)

        if n == 0:
            return self
        
        [other, *etc] = args

        if len(etc) == 0:
            if isinstance(other, Iterable):
                return lista.of(*self, *other)
            else:
                return lista.of(*self, other)
        
        return self.concat(other).concat(etc)
    
    @wraps(list.sort)
    def sort(self, **kwargs):
        self = lista(sorted(self, **kwargs))
        return self
    
    def __getitem__(self, index):
        if isinstance(index, slice):
            return lista(super().__getitem__(index))
        return super().__getitem__(index)
    
    @property
    def size(self):
        return lista._size(len(self))

class DictKeys(lista):
    def __call__(self):
        return self

class DictValues(lista):
    def __call__(self):
        return self
    
class DictItems(tuple):
    def __call__(self):
        return self

class dicc(dict):

    @staticmethod
    def pass_values(f):
        return lambda _,v: f(v)
    
    @staticmethod
    def pass_keys(f):
        return lambda k,_: f(k)

    def filter(self, f, over=None):
        """
        Versión eager de filter.
        """
        if over is None:
            over = 'items'
        
        match over:
            case 'keys':
                return self.filter_keys(f)
            case 'values':
                return self.filter_values(f)
            case 'items':
                return dicc({k: v for k, v in super().items() if f(k, v)})
            case _:
                raise ValueError('Invalid value for "by", must be one of "keys", "values" or "items"')

    def filter_keys(self, f):
        """
        Versión eager de filter para las claves del diccionario.
        """
        return dicc({k: v for k, v in super().items() if f(k, v)})
    
    def filter_values(self, f):
        """
        Versión eager de filter para los valores del diccionario.
        """
        return dicc({k: v for k, v in super().items() if f(k, v)})
    
    def map(self, f, over=None):
        """
        Versión eager de map.
        """

        if over is None:
            over = 'items'
            
        match over:
            case 'keys':
                return self.map_keys(f)
            case 'values':
                return self.map_values(f)
            case 'items':
                return lista(super().items()).map(lambda x: f(*x))
            case _:
                raise ValueError('Invalid value for "by", must be one of "keys", "values" or "items"')
    
    def map_values(self, f):
        """
        Versión eager de map para los valores del diccionario.
        """
        return dicc({k: f(k,v) for k, v in super().items()})
    
    def map_keys(self, f):
        """
        Versión eager de map para las claves del diccionario.
        """
        return dicc({f(k,v): v for k, v in super().items()})
    
    @property
    def keys(self):
        return DictKeys(super().keys())
    
    @property
    def values(self):
        return DictValues(super().values())
    
    @property
    def items(self):
        return DictItems(x for x in super().items())
    
    def __iter__(self) -> Iterator:
        return iter(self.items)
    

class bijection:
    @classmethod
    def from_list(cls, xs):
        if isinstance(xs, list):
            if all(map(lambda x: isinstance(x, tuple) and len(x) == 2, xs)):
                return cls(dict(xs))
            raise TypeError
        raise TypeError
                
    def __init__(self, d: None|dict = None):
        if d:
            self.forward = d
            self.backward = {v: k for k, v in d.items()}
        else:
            self.forward = {}
            self.backward = {}
    
    def __setitem__(self, key, value):
        self.forward[key] = value
        self.backward[value] = key
    
    def __getitem__(self, key):
        if key in self.forward:
            return self.forward[key]
        return self.backward[key]

    def get(self, key, default=None):
        if not (key in self.forward or key in self.backward):
            return default

        return self[key]
    
    def __contains__(self, key):
        return key in self.forward or key in self.backward
    
    def __delitem__(self, key):
        if key in self.forward:
            del self.backward[self.forward[key]]
            del self.forward[key]
        else:
            del self.forward[self.backward[key]]
            del self.backward[key]
    
    def __len__(self):
        return len(self.forward)
    
    def __iter__(self):
        return iter(self.forward.items())
    
    @staticmethod
    def __format__(k,v, align='left'):
        if align not in ('left', 'right'):
            raise ValueError(f"align must be 'left' or 'right', not {align}")
        
        str_left = getattr(str(k), 'ljust' if align == 'left' else 'rjust')
        str_right = getattr(str(v), 'rjust' if align == 'left' else 'ljust')
        
        return lambda l,r: f"{str_left(l)} <-> {str_right(r)}"
    
    def __repr__(self):
        output = 'bijection('
        offset = len(output)

        strings = [None for _ in self.forward]
        max_len_k = 0
        max_len_v = 0
        i = 0
        
        for k, v in self.forward.items():
            max_len_k = max(max_len_k, len(str(k)))
            max_len_v = max(max_len_v, len(str(v)))

            strings[i] = bijection.__format__(k,v, align='left')
            i += 1
        
        it = iter(map(lambda f: f(max_len_k, max_len_v), strings))
        output += next(it) + '\n'
        for s in it:
            output += (' '*offset) + s + '\n'
        output += (' '*(offset-1)) + ')'
        
        return output

# ======================================================================================================================

def get_type_name(_type: type) -> str:
    if _type is None:
        return 'None'
    elif hasattr(_type, '__name__'):
        return _type.__name__
    else:
        return str(_type)

class InitFromAnnotationsMeta(type):
    def __new__(cls, name, bases, dct):
        annotations = dct.get('__annotations__', {})

        dct['fields'] = [name for name in annotations]

        dct['not_null_fields'] = property(lambda self: [name for name in annotations if getattr(self, name) is not None])

        if annotations:
            parameters = [f'self'] + [f'{name}: {get_type_name(t)}' for name, t in annotations.items()]
            init_body = '\n    '.join([f'self.{name} = {name}' for name in annotations])

            # Default __init__
            exec_globals = globals().copy()
            default_init_definition = f"def default_init({', '.join(parameters)}):\n    {init_body}"
            exec(default_init_definition, exec_globals)
            default_init = exec_globals['default_init']

            # Is __init__ already defined?
            original_init = dct.get('__init__')

            if original_init:
                # Create a new __init__ that first calls the default and then the original
                def __init__(self, *args, **kwargs):
                    default_init(self, *args, **kwargs)
                    original_init(self, *args, **kwargs)
                dct['__init__'] = __init__
            else:
                dct['__init__'] = default_init

            # from_dict
            from_dict_body = ', '.join([f'data.get("{name}", None)' for name in annotations])
            from_dict_definition = f"@classmethod\ndef from_dict(cls, data):\n    return cls({from_dict_body})"
            exec(from_dict_definition, globals(), dct)

            # to_dict
            to_dict_body = ', '.join([f'"{name}": self.{name}' for name in annotations])
            to_dict_definition = f"def to_dict(self):\n    return {{{to_dict_body}}}"
            exec(to_dict_definition, globals(), dct)

            # Define equals method for class comparison
            #  def equals(self, other):
            #      if not isinstance(other, self.__class__):
            #          return False
            #      return all(
            #          (getattr(self, attr).equals(getattr(other, attr)) if isinstance(getattr(self, attr), AutoInit) else getattr(self, attr) == getattr(other, attr))
            #          for attr in annotations
            #      )
            #  dct['equals'] = equals

            # diff between objects. Similar to equals, but returns a list of the field names that differ.
            def diff(self, other):
                if not isinstance(other, self.__class__):
                    return list(annotations.keys())
                return [attr for attr in annotations if getattr(self, attr) != getattr(other, attr)]
            dct['diff'] = diff

            # Equals from diff == []
            def equals(self, other):
                return self.diff(other) == []
            dct['equals'] = equals

            if not '__str__' in dct:
                def __str__(self):
                    return f"{name}({', '.join([f'{name}={getattr(self, name)!r}' for name in annotations])})"
                dct['__str__'] = __str__

            if not '__repr__' in dct:
                def __repr__(self):
                    return f"{name}({', '.join([f'{name}={getattr(self, name)!r}' for name in annotations])})"
                dct['__repr__'] = __repr__

            @classmethod
            def from_other(cls, other, **kwargs):
                return cls.from_dict(**(other.to_dict()|kwargs))
            
            dct['from_other'] = from_other

            def update(self, other_or_dict, **kwargs):
                if isinstance(other_or_dict, dict):
                    keyset = set(other_or_dict)
                    annotation_set = set(annotations)
                    if not keyset.issubset(annotation_set) or not keyset == annotation_set:
                        raise ValueError(f"Invalid keys for {name}: {keyset - annotation_set}")
                    
                    for key, value in other_or_dict.items():
                        if value is not None:
                            setattr(self, key, value)
                    
                    return self
                else:
                    return self.update(other_or_dict.to_dict(), **kwargs)
            
            dct['update'] = update
            dct['__lshift__'] = update

        return super().__new__(cls, name, bases, dct)
    
class AutoInit(metaclass=InitFromAnnotationsMeta):
    pass

class Struct: ...
del Struct
Struct: type = AutoInit